from collections.abc import Mapping
from enum import Enum
from urllib.parse import unquote_plus

import dask.array as da
import numpy as np
import pandas
import xarray as xr


def gff3_parse_attributes(attributes_string):
    """Parse a string of GFF3 attributes ('key=value' pairs delimited by ';')
    and return a dictionary."""

    attributes = dict()
    fields = attributes_string.split(";")
    for f in fields:
        if "=" in f:
            key, value = f.split("=")
            key = unquote_plus(key).strip()
            value = unquote_plus(value.strip())
            attributes[key] = value
        elif len(f) > 0:
            # not strictly kosher, treat as a flag
            attributes[unquote_plus(f).strip()] = True
    return attributes


gff3_cols = (
    "seqid",
    "source",
    "type",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attributes",
)


def read_gff3(buf, compression="gzip"):

    # read as dataframe
    df = pandas.read_csv(
        buf,
        sep="\t",
        comment="#",
        names=gff3_cols,
        na_values=["", "."],
        compression=compression,
    )

    # parse attributes
    df["attributes"] = df["attributes"].apply(gff3_parse_attributes)

    return df


def unpack_gff3_attributes(df, attributes):

    df = df.copy()

    if attributes == "*":
        # discover all attribute keys
        attributes = set()
        for a in df["attributes"]:
            attributes.update(a.keys())
        attributes = sorted(attributes)

    # unpack attributes into columns
    for key in attributes:
        df[key] = df["attributes"].apply(lambda v: v.get(key, np.nan))
    del df["attributes"]

    return df


class SafeStore(Mapping):
    def __init__(self, store):
        self.store = store

    def __getitem__(self, key):
        try:
            return self.store[key]
        except KeyError as e:
            # always raise a runtime error to ensure zarr propagates the exception
            raise RuntimeError(e)

    def __contains__(self, key):
        return key in self.store

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)


class SiteClass(Enum):
    UPSTREAM = 1
    DOWNSTREAM = 2
    UTR5 = 3
    UTR3 = 4
    CDS_FIRST = 5
    CDS_MID = 6
    CDS_LAST = 7
    INTRON_FIRST = 8
    INTRON_MID = 9
    INTRON_LAST = 10


def from_zarr(z, inline_array, chunks="auto"):
    """Utility function for turning a zarr array into a dask array.

    N.B., dask does have it's own from_zarr() function but we roll
    our own here to get a little more control.

    """
    if chunks == "native":
        chunks = z.chunks
    kwargs = dict(chunks=chunks, fancy=False, lock=False, inline_array=inline_array)
    try:
        d = da.from_array(z, **kwargs)
    except TypeError:
        # only later versions of dask support inline_array argument
        del kwargs["inline_array"]
        d = da.from_array(z, **kwargs)
    return d


def dask_compress_dataset(ds, indexer, dim):
    """Temporary workaround for memory issues when attempting to
    index an xarray dataset with a Boolean array.

    See also: https://github.com/pydata/xarray/issues/5054

    Parameters
    ----------
    ds : xarray.Dataset
    indexer : str
    dim : str

    Returns
    -------
    xarray.Dataset

    """
    if isinstance(indexer, str):
        indexer = ds[indexer].data

    # sanity checks
    assert isinstance(indexer, da.Array)
    assert indexer.ndim == 1
    assert indexer.dtype == bool
    assert indexer.shape[0] == ds.dims[dim]

    coords = dict()
    for k in ds.coords:
        a = ds[k]
        v = _dask_compress_dataarray(a, indexer, dim)
        coords[k] = (a.dims, v)

    data_vars = dict()
    for k in ds.data_vars:
        a = ds[k]
        v = _dask_compress_dataarray(a, indexer, dim)
        data_vars[k] = (a.dims, v)

    attrs = ds.attrs.copy()

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


def _dask_compress_dataarray(a, indexer, dim):
    try:
        # find the axis for the given dimension
        axis = a.dims.index(dim)

    except ValueError:
        # array doesn't have the given dimension, return as-is
        v = a.data

    else:
        # apply the indexing operation
        v = dask_compress(indexer, a.data, axis)

    return v


def dask_compress(indexer, data, axis):
    """Wrapper for dask.array.compress() which computes chunk sizes faster."""

    # sanity checks
    assert isinstance(data, da.Array)
    assert isinstance(indexer, da.Array)
    assert isinstance(axis, int)
    assert indexer.shape[0] == data.shape[axis]
    old_chunks = data.chunks
    axis_old_chunks = old_chunks[axis]
    axis_n_chunks = len(axis_old_chunks)

    # apply the indexing operation
    v = da.compress(indexer, data, axis=axis)

    # need to compute chunks sizes in order to know dimension sizes;
    # would normally do v.compute_chunk_sizes() but that is slow for
    # multidimensional arrays, so hack something more efficient

    axis_new_chunks = tuple(
        indexer.rechunk((axis_old_chunks,))
        .map_blocks(
            lambda b: np.sum(b, keepdims=True),
            chunks=((1,) * axis_n_chunks,),
        )
        .compute()
    )
    new_chunks = tuple(
        [axis_new_chunks if i == axis else c for i, c in enumerate(old_chunks)]
    )
    v._chunks = new_chunks

    return v



from __future__ import division, print_function

import collections
import operator

import petl as etl
from Bio.Seq import Seq


class Annotator(object):
    def __init__(self, genome, geneset):
        """
        An annotator.

        Parameters
        ----------

        genome : zarr hierarchy
            Reference genome.
        geneset : pandas dataframe
            Dataframe with genome annotations.

        """

        # store initialisation parameters
        self._genome = genome
        self._genome_cache = dict()
        # when debugging snp effects unhash seqid and add .eq("seqid", seqid) parameter to tbl_features
        # and seqid to __init__
        # self._seqid = seqid

        # setup access to GFF3 as a petl table
        # TODO at some point we'd like to refactor this module to read directly from pandas
        tbl_features = (
            etl.fromdataframe(geneset)
            .rename({"ID": "feature_id", "Parent": "parent_id", "end": "stop"})
            .select(lambda row: (row.stop - row.start) > 0)
        )
        self._tbl_features = tbl_features.cache()

        # index features by ID
        self._idx_feature_id = self._tbl_features.recordlookupone("feature_id")

        # index features by parent ID
        self._idx_parent_id = self._tbl_features.recordlookup("parent_id")

        # index features by genomic location
        self._idx_location = self._tbl_features.facetintervalrecordlookup(
            "seqid", "start", "stop", include_stop=True
        )

    def get_feature(self, feature_id):
        return self._idx_feature_id[feature_id]

    def get_children(self, feature_id):
        return self._idx_parent_id[feature_id]

    def find(self, chrom, start, stop):
        return self._idx_location[chrom].search(start, stop)

    def get_ref_seq(self, chrom, start, stop):
        """Accepts 1-based coords."""
        try:
            seq = self._genome_cache[chrom]
        except KeyError:
            seq = self._genome[chrom][:]
            self._genome_cache[chrom] = seq
        ref_seq = seq[start - 1 : stop]
        ref_seq = ref_seq.tobytes().decode()
        return ref_seq

    def get_ref_allele_coords(self, chrom, pos, ref):

        # N.B., use one-based inclusive coordinate system (like GFF3) throughout
        ref_start = pos
        ref_stop = pos + len(ref) - 1

        # check the reference allele matches the reference sequence
        ref_seq = self.get_ref_seq(chrom, ref_start, ref_stop).lower()
        assert ref_seq == ref.lower(), (
            "reference allele does not match reference sequence, "
            "expected %r, found %r" % (ref_seq, ref.lower())
        )

        return ref_start, ref_stop


VariantEffect = collections.namedtuple(
    "VariantEffect",
    (
        "effect",
        "impact",
        "chrom",
        "pos",
        "ref",
        "alt",
        "vlen",
        "ref_start",
        "ref_stop",
        "gene_id",
        "gene_start",
        "gene_stop",
        "gene_strand",
        "transcript_id",
        "transcript_start",
        "transcript_stop",
        "transcript_strand",
        "cds_id",
        "cds_start",
        "cds_stop",
        "cds_strand",
        "intron_start",
        "intron_stop",
        "intron_5prime_dist",
        "intron_3prime_dist",
        "intron_exon_5prime",
        "intron_exon_3prime",
        "ref_cds_start",
        "ref_cds_stop",
        "ref_intron_start",
        "ref_intron_stop",
        "ref_start_phase",
        "ref_codon",
        "alt_codon",
        "codon_change",
        "aa_pos",
        "ref_aa",
        "alt_aa",
        "aa_change",
    ),
)
null_effect = VariantEffect(*([None] * len(VariantEffect._fields)))


def get_effects(
    annotator,
    chrom,
    pos,
    ref,
    alt,
    gff3_gene_types={"gene", "pseudogene"},
    gff3_transcript_types={"mRNA", "rRNA", "pseudogenic_transcript"},
    gff3_cds_types={"CDS", "pseudogenic_exon"},
    transcript_ids=None,
):
    """TODO

    Parameters
    ----------

    gff3_gene_types : list of strings, optional
        Feature types to consider as genes.
    gff3_transcript_types : list of strings, optional
        Feature types to consider as transcripts.
    gff3_cds_types : list of strings, optional
        Feature types to consider as coding sequences.

    Returns
    -------

    A `VariantEffect` generator.

    """

    # ensure types and case
    ref = str(ref).upper()
    alt = str(alt).upper()

    # obtain start and stop coordinates of the reference allele
    ref_start, ref_stop = annotator.get_ref_allele_coords(chrom, pos, ref)

    # setup common effect parameters
    base_effect = null_effect._replace(
        chrom=chrom,
        pos=pos,
        ref=ref,
        alt=alt,
        vlen=len(alt) - len(ref),
        ref_start=ref_start,
        ref_stop=ref_stop,
    )

    # find overlapping genome features
    features = annotator.find(chrom, ref_start, ref_stop)

    # filter to find overlapping genes
    genes = [f for f in features if f.type in gff3_gene_types]

    if not genes:
        for effect in _get_intergenic_effects(annotator, base_effect):
            yield effect

    else:
        for gene in genes:
            for effect in _get_gene_effects(
                annotator,
                base_effect,
                gene,
                gff3_transcript_types,
                gff3_cds_types,
                transcript_ids,
            ):
                yield effect


# add as method
Annotator.get_effects = get_effects


def _get_intergenic_effects(annotator, base_effect):

    # TODO
    # UPSTREAM and DOWNSTREAM

    # the variant is in an intergenic region
    effect = base_effect._replace(effect="INTERGENIC", impact="MODIFIER")
    yield effect


def _get_gene_effects(
    annotator, base_effect, gene, gff3_transcript_types, gff3_cds_types, transcript_ids
):

    # setup common effect parameters
    base_effect = base_effect._replace(
        gene_id=gene.feature_id,
        gene_start=gene.start,
        gene_stop=gene.stop,
        gene_strand=gene.strand,
    )

    # obtain transcripts that are children of the current gene
    transcripts = [
        t
        for t in annotator.get_children(gene.feature_id)
        if t.type in gff3_transcript_types
    ]

    if not transcript_ids and not transcripts:

        # the variant hits a gene, but no transcripts within the gene
        effect = base_effect._replace(effect="INTRAGENIC", impact="MODIFIER")
        yield effect

    else:

        # optionally filter to user-specified transcripts
        if transcript_ids:
            transcripts = [t for t in transcripts if t.feature_id in transcript_ids]

        for transcript in transcripts:
            for effect in _get_transcript_effects(
                annotator, base_effect, transcript, gff3_cds_types
            ):
                yield effect


def _get_transcript_effects(annotator, base_effect, transcript, gff3_cds_types):

    # setup common effect parameters
    base_effect = base_effect._replace(
        transcript_id=transcript.feature_id,
        transcript_start=transcript.start,
        transcript_stop=transcript.stop,
        transcript_strand=transcript.strand,
    )

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    transcript_start = transcript.start
    transcript_stop = transcript.stop

    # compare start and stop of reference allele to start
    # and stop of current transcript

    if ref_stop < transcript_start:

        # TODO
        # variant hits a gene but misses the current transcript, falling
        # upstream
        effect = base_effect._replace(effect="TODO")
        yield effect

    elif ref_start > transcript_stop:

        # TODO
        # variant hits a gene but misses the current transcript, falling
        # downstream
        effect = base_effect._replace(effect="TODO")
        yield effect

    elif ref_start < transcript_start <= ref_stop <= transcript_stop:

        # TODO
        # reference allele overhangs the start of the current transcript
        effect = base_effect._replace(effect="TODO")
        yield effect

    elif transcript_start <= ref_start <= transcript_stop < ref_stop:

        # TODO
        # reference allele overhangs the end of the current transcript
        effect = base_effect._replace(effect="TODO")
        yield effect

    elif ref_start < transcript_start <= transcript_stop < ref_stop:

        # TODO
        # reference allele entirely overlaps the current transcript and
        # overhangs at both ends
        effect = base_effect._replace(effect="TODO")
        yield effect

    else:

        # reference allele falls within current transcript
        assert transcript_start <= ref_start <= ref_stop <= transcript_stop
        for effect in _get_within_transcript_effects(
            annotator, base_effect, transcript, gff3_cds_types
        ):
            yield effect


def _get_within_transcript_effects(annotator, base_effect, transcript, gff3_cds_types):

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop

    # obtain coding sequences that are children of the current transcript
    cdss = sorted(
        [
            f
            for f in annotator.get_children(transcript.feature_id)
            if f.type in gff3_cds_types
        ],
        key=lambda v: v.start,
    )

    exons = sorted(
        [f for f in annotator.get_children(transcript.feature_id) if f.type == "exon"],
        key=lambda v: v.start,
    )

    utr5 = sorted(
        [
            f
            for f in annotator.get_children(transcript.feature_id)
            if f.type == "five_prime_UTR"
        ],
        key=lambda v: v.start,
    )

    utr3 = sorted(
        [
            f
            for f in annotator.get_children(transcript.feature_id)
            if f.type == "three_prime_UTR"
        ],
        key=lambda v: v.start,
    )

    # derive introns, assuming between exons
    introns = [(x.stop + 1, y.start - 1) for x, y in zip(exons[:-1], exons[1:])]

    # introns_5utr = [(x.stop + 1, y.start - 1) for x, y in zip(utr5[:-1], utr5[1:])]
    #
    # # derive introns, assuming between CDSs
    # introns = [(x.stop + 1, y.start - 1) for x, y in zip(cdss[:-1], cdss[1:])]

    # if not cdss:
    #
    #     # TODO
    #     # the variant hits a transcript, but there are no CDSs within the
    #     # transcript
    #     effect = base_effect._replace(effect="TODO")
    #     yield effect

    # find coding sequence that overlaps the reference allele
    overlapping_cdss = [
        cds for cds in cdss if cds.start <= ref_stop and cds.stop >= ref_start
    ]

    overlapping_introns = [
        (start, stop)
        for (start, stop) in introns
        if start <= ref_stop and stop >= ref_start
    ]

    # overlapping_5utr_introns = [
    #     (start, stop)
    #     for (start, stop) in introns_5utr
    #     if start <= ref_stop and stop >= ref_start
    # ]

    overlapping_utr5 = [x for x in utr5 if x.start <= ref_stop and x.stop >= ref_start]

    overlapping_utr3 = [x for x in utr3 if x.start <= ref_stop and x.stop >= ref_start]

    # CDS effects

    if overlapping_cdss:

        if len(overlapping_cdss) > 1:

            # TODO
            # variant overlaps more than one exon
            effect = base_effect._replace(effect="TODO")
            yield effect

        else:

            # variant overlaps a single exon
            assert len(overlapping_cdss) == 1
            cds = overlapping_cdss[0]

            yield _get_cds_effect(annotator, base_effect, cds, cdss)

    # intron effects

    if overlapping_introns:

        if len(overlapping_introns) > 1:

            # TODO
            # variant overlaps more than one intron
            effect = base_effect._replace(effect="TODO")
            yield effect

        else:

            # variant overlaps a single intron
            assert len(overlapping_introns) == 1
            intron = overlapping_introns[0]

            yield _get_intron_effect(annotator, base_effect, intron, exons)

    # if overlapping_5utr_introns:
    #
    #     if len(overlapping_5utr_introns) > 1:
    #
    #         # TODO
    #         # variant overlaps more than one intron
    #         effect = base_effect._replace(effect="TODO")
    #         yield effect
    #
    #     else:
    #
    #         # variant overlaps a single intron in 5' utr
    #         assert len(overlapping_5utr_introns) == 1
    #         # TODO why doesnt this work
    #         # intron = overlapping_5utr_introns[0]
    #         # yield _get_intron_effect(annotator, base_effect, intron, cdss)
    #         effect = base_effect._replace(effect="INTRONIC")
    #         yield effect

    if overlapping_utr5:

        if len(overlapping_utr5) > 1:

            # TODO
            # variant overlaps more than one 5'UTR
            effect = base_effect._replace(effect="TODO")
            yield effect

        else:

            # variant overlaps a single 5 prime UTR
            assert len(overlapping_utr5) == 1
            utr5 = overlapping_utr5[0]

            effect = base_effect._replace(effect="FIVE_PRIME_UTR", impact="LOW")
            yield effect

    if overlapping_utr3:

        if len(overlapping_utr3) > 1:

            # TODO
            # variant overlaps more than one 3'UTR
            effect = base_effect._replace(effect="TODO")
            yield effect

        else:

            # variant overlaps a single 3 prime UTR
            assert len(overlapping_utr3) == 1
            utr3 = overlapping_utr3[0]

            effect = base_effect._replace(effect="THREE_PRIME_UTR", impact="LOW")
            yield effect

    # if none of the above - #
    if (
        (not overlapping_cdss)
        and (not overlapping_introns)
        and (not overlapping_utr5)
        and (not overlapping_utr3)
    ):
        effect = base_effect._replace(effect="INTRAGENIC", impact="LOW")
        yield effect


def _get_cds_effect(annotator, base_effect, cds, cdss):

    # setup common effect parameters
    base_effect = base_effect._replace(
        cds_id=cds.feature_id,
        cds_start=cds.start,
        cds_stop=cds.stop,
        cds_strand=cds.strand,
    )

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    cds_start = cds.start
    cds_stop = cds.stop

    if ref_start < cds_start <= ref_stop <= cds_stop:

        # TODO
        # reference allele overhangs the start of the current exon
        effect = base_effect._replace(effect="TODO")
        return effect

    elif cds_start <= ref_start <= cds_stop < ref_stop:

        # TODO
        # reference allele overhangs the end of the current transcript
        effect = base_effect._replace(effect="TODO")
        return effect

    elif ref_start < cds_start <= cds_stop < ref_stop:

        # TODO
        # reference allele entirely overlaps the current exon and
        # overhangs at both ends
        effect = base_effect._replace(effect="TODO")
        return effect

    else:

        # reference allele falls within current transcript
        assert cds_start <= ref_start <= ref_stop <= cds_stop
        return _get_within_cds_effect(annotator, base_effect, cds, cdss)


def _get_within_cds_effect(annotator, base_effect, cds, cdss):

    # convenience
    chrom = base_effect.chrom
    pos = base_effect.pos
    ref = base_effect.ref
    alt = base_effect.alt
    strand = base_effect.cds_strand

    # obtain amino acid change
    (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
        aa_pos,
        ref_aa,
        alt_aa,
    ) = _get_aa_change(annotator, chrom, pos, ref, alt, cds, cdss)

    # setup common effect parameters
    base_effect = base_effect._replace(
        ref_cds_start=ref_cds_start,
        ref_cds_stop=ref_cds_stop,
        ref_start_phase=ref_start_phase,
        ref_codon=ref_codon,
        alt_codon=alt_codon,
        codon_change="%s/%s" % (ref_codon, alt_codon),
        aa_pos=aa_pos,
        ref_aa=ref_aa,
        alt_aa=alt_aa,
        aa_change="%s%s%s" % (ref_aa, aa_pos, alt_aa),
    )

    if len(ref) == 1 and len(alt) == 1:

        # SNPs

        if ref_aa == alt_aa:

            # TODO SYNONYMOUS_START and SYNONYMOUS_STOP

            # variant causes a codon that produces the same amino acid
            # e.g.: Ttg/Ctg, L/L
            effect = base_effect._replace(effect="SYNONYMOUS_CODING", impact="LOW")

        elif ref_aa == "M" and ref_cds_start == 0:

            # variant causes start codon to be mutated into a non-start codon.
            # e.g.: aTg/aGg, M/R
            effect = base_effect._replace(effect="START_LOST", impact="HIGH")

        elif ref_aa == "*":

            # variant causes stop codon to be mutated into a non-stop codon
            # e.g.: Tga/Cga, */R
            effect = base_effect._replace(effect="STOP_LOST", impact="HIGH")

        elif alt_aa == "*":

            # variant causes a STOP codon e.g.: Cag/Tag, Q/*
            effect = base_effect._replace(effect="STOP_GAINED", impact="HIGH")

        else:

            # TODO NON_SYNONYMOUS_START and NON_SYNONYMOUS_STOP

            # variant causes a codon that produces a different amino acid
            # e.g.: Tgg/Cgg, W/R
            effect = base_effect._replace(
                effect="NON_SYNONYMOUS_CODING", impact="MODERATE"
            )

    else:

        # INDELs and MNPs

        if (len(alt) - len(ref)) % 3:

            # N.B., this case covers both simple INDELs and complex
            # polymorphisms

            # insertion or deletion causes a frame shift
            # e.g.: An indel size is not multple of 3
            effect = base_effect._replace(effect="FRAME_SHIFT", impact="HIGH")

        elif len(ref) == 1 and len(ref) < len(alt):

            # simple insertions

            # figure out if there has been a codon change or not
            is_codon_changed = (strand == "+" and ref_aa[0] != alt_aa[0]) or (
                strand == "-" and ref_aa[-1] != alt_aa[-1]
            )

            if is_codon_changed:

                # one codon is changed and one or many codons are inserted
                # e.g.: An insert of size multiple of three, not at codon
                # boundary
                effect = base_effect._replace(
                    effect="CODON_CHANGE_PLUS_CODON_INSERTION", impact="MODERATE"
                )

            else:

                # one or many codons are inserted
                # e.g.: An insert multiple of three in a codon boundary
                effect = base_effect._replace(
                    effect="CODON_INSERTION", impact="MODERATE"
                )

        elif len(alt) == 1 and len(ref) > len(alt):

            # simple deletions

            # figure out if there has been a codon change or not
            is_codon_changed = (strand == "+" and ref_aa[0] != alt_aa[0]) or (
                strand == "-" and ref_aa[-1] != alt_aa[-1]
            )

            if is_codon_changed:

                # one codon is changed and one or many codons are deleted
                # e.g.: A deletion of size multiple of three, not at codon
                # boundary
                effect = base_effect._replace(
                    effect="CODON_CHANGE_PLUS_CODON_DELETION", impact="MODERATE"
                )

            else:

                # one or many codons are deleted
                # e.g.: A deletions multiple of three in a codon boundary
                effect = base_effect._replace(
                    effect="CODON_DELETION", impact="MODERATE"
                )

        elif len(ref) == len(alt):

            # MNPs
            effect = base_effect._replace(effect="CODON_CHANGE", impact="MODERATE")

        else:

            # TODO in-frame complex variation (MNP + INDEL)
            effect = base_effect._replace(effect="TODO")

    return effect


def _get_aa_change(annotator, chrom, pos, ref, alt, cds, cdss):

    # obtain codon change
    (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
    ) = _get_codon_change(annotator, chrom, pos, ref, alt, cds, cdss)

    # translate codon change to amino acid change
    ref_aa = str(Seq(ref_codon).translate())
    alt_aa = str(Seq(alt_codon).translate())
    aa_pos = (ref_cds_start // 3) + 1

    return (
        ref_cds_start,
        ref_cds_stop,
        ref_start_phase,
        ref_codon,
        alt_codon,
        aa_pos,
        ref_aa,
        alt_aa,
    )


def _get_codon_change(annotator, chrom, pos, ref, alt, cds, cdss):

    # obtain reference allele coords relative to coding sequence
    ref_start, ref_stop = annotator.get_ref_allele_coords(chrom, pos, ref)
    ref_cds_start, ref_cds_stop = _get_coding_position(ref_start, ref_stop, cds, cdss)

    # calculate position of reference allele start within codon
    ref_start_phase = ref_cds_start % 3

    if cds.strand == "+":

        # obtain any previous nucleotides to complete the first codon
        prefix = annotator.get_ref_seq(
            chrom=chrom, start=ref_start - ref_start_phase, stop=ref_start - 1
        ).lower()

        # begin constructing reference and alternate codon sequences
        ref_codon = prefix + ref
        alt_codon = prefix + alt

        # obtain any subsequence nucleotides to complete the last codon
        if len(ref_codon) % 3:
            ref_stop_phase = len(ref_codon) % 3
            suffix = annotator.get_ref_seq(
                chrom=chrom, start=ref_stop + 1, stop=ref_stop + 3 - ref_stop_phase
            )
            suffix = str(suffix).lower()
            ref_codon += suffix

        if len(alt_codon) % 3:
            alt_stop_phase = len(alt_codon) % 3
            suffix = annotator.get_ref_seq(
                chrom=chrom, start=ref_stop + 1, stop=ref_stop + 3 - alt_stop_phase
            ).lower()
            alt_codon += suffix

    else:

        # N.B., we are on the reverse strand, so position reported for
        # variant is actually position at the *end* of the reference allele
        # which is particularly important for deletions

        # we will construct everything for the forward strand (i.e., back-to-
        # front) then take reverse complement afterwards at the end of this
        # code block

        # obtain any previous nucleotides to complete the first codon
        prefix = annotator.get_ref_seq(
            chrom=chrom, start=ref_stop + 1, stop=ref_stop + ref_start_phase
        ).lower()

        # begin constructing reference and alternate codon sequences
        ref_codon = ref + prefix
        alt_codon = alt + prefix

        # obtain any subsequence nucleotides to complete the last codon
        if len(ref_codon) % 3:
            ref_stop_phase = len(ref_codon) % 3
            suffix = annotator.get_ref_seq(
                chrom=chrom, start=ref_start - 3 + ref_stop_phase, stop=ref_start - 1
            ).lower()
            ref_codon = suffix + ref_codon

        if len(alt_codon) % 3:
            alt_stop_phase = len(alt_codon) % 3
            suffix = annotator.get_ref_seq(
                chrom=chrom, start=ref_start - 3 + alt_stop_phase, stop=ref_start - 1
            ).lower()
            alt_codon = suffix + alt_codon

        # take reverse complement
        ref_codon = str(Seq(ref_codon).reverse_complement())
        alt_codon = str(Seq(alt_codon).reverse_complement())

    return ref_cds_start, ref_cds_stop, ref_start_phase, ref_codon, alt_codon


def _get_coding_position(ref_start, ref_stop, cds, cdss):
    # print('_get_coding_position', ref_start, ref_stop, cds, len(cdss))

    if cds.strand == "+":

        # sort exons
        cdss = sorted(cdss, key=operator.itemgetter("start"))

        # find index of overlapping exons in all exons
        cds_index = [f.start for f in cdss].index(cds.start)
        # print('_get_coding_position (+) cds_index', cds_index)

        # find offset
        offset = sum([f.stop - f.start + 1 for f in cdss[:cds_index]])
        # print('_get_coding_position (+) offset', offset)

        # find ref cds position
        ref_cds_start = offset + (ref_start - cds.start)
        ref_cds_stop = offset + (ref_stop - cds.start)

    else:

        # sort exons (backwards this time)
        cdss = sorted(cdss, key=operator.itemgetter("stop"), reverse=True)

        # find index of overlapping exons in all exons
        cds_index = [f.stop for f in cdss].index(cds.stop)
        # print('_get_coding_position (-) cds_index', cds_index)

        # find offset
        offset = sum([cds.stop - cds.start + 1 for cds in cdss[:cds_index]])
        # print('_get_coding_position (-) offset', offset)

        # find ref cds position
        ref_cds_start = offset + (cds.stop - ref_stop)
        ref_cds_stop = offset + (cds.stop - ref_start)

    # print('_get_coding_position return', ref_cds_start, ref_cds_stop)
    return ref_cds_start, ref_cds_stop


def _get_intron_effect(annotator, base_effect, intron, exons):

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    intron_start, intron_stop = intron

    if ref_start < intron_start <= ref_stop <= intron_stop:

        # TODO
        # reference allele overhangs the start of the current intron
        effect = base_effect._replace(effect="TODO")
        return effect

    elif intron_start <= ref_start <= intron_stop < ref_stop:

        # TODO
        # reference allele overhangs the end of the current intron
        effect = base_effect._replace(effect="TODO")
        return effect

    elif ref_start < intron_start <= intron_stop < ref_stop:

        # TODO
        # reference allele entirely overlaps the current intron and
        # overhangs at both ends
        effect = base_effect._replace(effect="TODO")
        return effect

    else:

        # reference allele falls within current intron
        assert intron_start <= ref_start <= ref_stop <= intron_stop
        return _get_within_intron_effect(annotator, base_effect, intron, exons)


def _get_within_intron_effect(annotator, base_effect, intron, exons):

    # convenience
    ref_start = base_effect.ref_start
    ref_stop = base_effect.ref_stop
    ref = base_effect.ref
    alt = base_effect.alt
    intron_start, intron_stop = intron
    strand = base_effect.gene_strand
    if strand == "+":
        intron_5prime_dist = ref_start - (intron_start - 1)
        intron_3prime_dist = ref_stop - (intron_stop + 1)
        intron_exon_5prime = [
            exon.feature_id for exon in exons if exon.stop == intron_start - 1
        ][0]
        intron_exon_3prime = [
            exon.feature_id for exon in exons if exon.start == intron_stop + 1
        ][0]
    else:
        intron_5prime_dist = (intron_stop + 1) - ref_stop
        intron_3prime_dist = (intron_start - 1) - ref_start
        intron_exon_3prime = [
            exon.feature_id for exon in exons if exon.stop == intron_start - 1
        ][0]
        intron_exon_5prime = [
            exon.feature_id for exon in exons if exon.start == intron_stop + 1
        ][0]

    # setup common effect parameters
    base_effect = base_effect._replace(
        intron_start=intron_start,
        intron_stop=intron_stop,
        ref_intron_start=ref_start - intron_start,
        ref_intron_stop=ref_stop - intron_start,
        intron_5prime_dist=intron_5prime_dist,
        intron_3prime_dist=intron_3prime_dist,
        intron_exon_5prime=intron_exon_5prime,
        intron_exon_3prime=intron_exon_3prime,
    )

    intron_min_dist = min(intron_5prime_dist, -intron_3prime_dist)

    if len(ref) == 1 and len(alt) == 1:

        # SNPs

        if intron_min_dist <= 2:
            # splice site variation
            effect = base_effect._replace(effect="SPLICE_CORE", impact="HIGH")

        elif intron_min_dist <= 7:

            # splice site variation
            effect = base_effect._replace(effect="SPLICE_REGION", impact="MODERATE")

        else:

            # intron modifier
            effect = base_effect._replace(effect="INTRONIC", impact="MODIFIER")

    else:

        # TODO
        # INDELs and MNPs
        effect = base_effect._replace(effect="TODO")

    return effect



import allel
import dask.array as da
import numpy as np
import pandas
import xarray
import zarr
from fsspec.core import url_to_fs
from fsspec.mapping import FSMap

from . import veff
from .util import (
    SafeStore,
    dask_compress,
    dask_compress_dataset,
    from_zarr,
    read_gff3,
    unpack_gff3_attributes,
)

public_releases = ("v3",)
gff3_path = (
    "reference/genome/agamp4/Anopheles-gambiae-PEST_BASEFEATURES_AgamP4.12.gff3.gz"
)


DIM_VARIANT = "variants"
DIM_ALLELE = "alleles"
DIM_SAMPLE = "samples"
DIM_PLOIDY = "ploidy"


class Ag3:
    """Provides access to data from Ag 3 releases.

    Parameters
    ----------
    url : str
        Base path to data. Give "gs://vo_agam_release/" to use Google Cloud Storage,
        or a local path on your file system if data have been downloaded.
    **kwargs
        Passed through to fsspec when setting up file system access.

    Examples
    --------
    Access data from Google Cloud Storage:

        >>> import malariagen_data
        >>> ag3 = malariagen_data.Ag3("gs://vo_agam_release/")

    Access data downloaded to a local file system:

        >>> ag3 = malariagen_data.Ag3("/local/path/to/vo_agam_release/")

    """

    public_releases = ("v3",)
    contigs = ("2R", "2L", "3R", "3L", "X")

    def __init__(self, url, **kwargs):

        # special case Google Cloud Storage, use anonymous access, avoids a delay
        if url.startswith("gs://") or url.startswith("gcs://"):
            kwargs["token"] = "anon"
        elif "gs://" in url:
            # chained URL
            kwargs["gs"] = dict(token="anon")
        elif "gcs://" in url:
            # chained URL
            kwargs["gcs"] = dict(token="anon")

        # process the url using fsspec
        self._pre = kwargs.pop("pre", False)
        fs, path = url_to_fs(url, **kwargs)
        self._fs = fs
        # path compatibility, fsspec/gcsfs behaviour varies between version
        while path.endswith("/"):
            path = path[:-1]
        self._path = path

        # setup caches
        self._cache_releases = None
        self._cache_sample_sets = dict()
        self._cache_general_metadata = dict()
        self._cache_species_calls = dict()
        self._cache_site_filters = dict()
        self._cache_snp_sites = None
        self._cache_snp_genotypes = dict()
        self._cache_genome = None
        self._cache_annotator = None
        self._cache_geneset = dict()
        self._cache_cross_metadata = None
        self._cache_site_annotations = None

    @property
    def releases(self):
        if self._cache_releases is None:
            if self._pre:
                # discover which releases are available
                sub_dirs = [p.split("/")[-1] for p in self._fs.ls(self._path)]
                releases = sorted([d for d in sub_dirs if d.startswith("v3")])
                if len(releases) == 0:
                    raise ValueError("No releases found.")
                self._cache_releases = releases
            else:
                self._cache_releases = public_releases
        return self._cache_releases

    def sample_sets(self, release="v3"):
        """Access the manifest of sample sets.

        Parameters
        ----------
        release : str
            Release identifier. Give "v3" to access the Ag1000G phase 3 data release.

        Returns
        -------
        df : pandas.DataFrame

        """

        if release not in self.releases:
            raise ValueError(f"Release not available: {release!r}")

        try:
            return self._cache_sample_sets[release]

        except KeyError:
            path = f"{self._path}/{release}/manifest.tsv"
            with self._fs.open(path) as f:
                df = pandas.read_csv(f, sep="\t", na_values="")
            df["release"] = release
            self._cache_sample_sets[release] = df
            return df

    @property
    def v3_wild(self):
        return [
            x
            for x in self.sample_sets(release="v3")["sample_set"].tolist()
            if x != "AG1000G-X"
        ]

    def _lookup_release(self, *, sample_set):
        # find which release this sample set was included in
        for release in self.releases:
            df_sample_sets = self.sample_sets(release=release)
            if sample_set in df_sample_sets["sample_set"].tolist():
                return release
        raise ValueError(f"No release found for sample set {sample_set!r}")

    def _read_general_metadata(self, *, sample_set):
        """Read metadata for a single sample set."""
        try:
            return self._cache_general_metadata[sample_set]
        except KeyError:
            release = self._lookup_release(sample_set=sample_set)
            path = (
                f"{self._path}/{release}/metadata/general/{sample_set}/samples.meta.csv"
            )
            with self._fs.open(path) as f:
                df = pandas.read_csv(f, na_values="")

            # add a couple of columns for convenience
            df["sample_set"] = sample_set
            df["release"] = release

            self._cache_general_metadata[sample_set] = df
            return df

    def _read_species_calls(self, *, sample_set, analysis, method):
        """Read species calls for a single sample set."""
        key = (sample_set, analysis, method)
        try:
            return self._cache_species_calls[key]
        except KeyError:
            release = self._lookup_release(sample_set=sample_set)
            path = (
                f"{self._path}/{release}/metadata/species_calls_{analysis}"
                f"/{sample_set}/samples.species_{method}.csv"
            )
            with self._fs.open(path) as f:
                df = pandas.read_csv(
                    f,
                    na_values="",
                    # ensure correct dtype even where all values are missing
                    dtype={
                        "species_gambcolu_arabiensis": object,
                        "species_gambiae_coluzzii": object,
                    },
                )

            # add a single species call column, for convenience
            def consolidate_species(s):
                species_gambcolu_arabiensis = s["species_gambcolu_arabiensis"]
                species_gambiae_coluzzii = s["species_gambiae_coluzzii"]
                if species_gambcolu_arabiensis == "arabiensis":
                    return "arabiensis"
                elif species_gambcolu_arabiensis == "intermediate":
                    return "intermediate_arabiensis_gambiae"
                elif species_gambcolu_arabiensis == "gamb_colu":
                    # look at gambiae_vs_coluzzii
                    if species_gambiae_coluzzii == "gambiae":
                        return "gambiae"
                    elif species_gambiae_coluzzii == "coluzzii":
                        return "coluzzii"
                    elif species_gambiae_coluzzii == "intermediate":
                        return "intermediate_gambiae_coluzzii"
                else:
                    # some individuals, e.g., crosses, have a missing species call
                    return np.nan

            df["species"] = df.apply(consolidate_species, axis=1)

            self._cache_species_calls[key] = df
            return df

    def _prep_sample_sets_arg(self, *, sample_sets):
        if sample_sets == "v3_wild":
            # convenience, special case to exclude crosses
            sample_sets = self.v3_wild

        elif isinstance(sample_sets, str) and sample_sets.startswith("v3"):
            # convenience, can use a release identifier to denote all sample sets
            # in a release
            sample_sets = self.sample_sets(release=sample_sets)["sample_set"].tolist()

        if not isinstance(sample_sets, (str, list, tuple)):
            raise TypeError(f"Invalid sample_sets: {sample_sets!r}")

        return sample_sets

    def species_calls(self, sample_sets="v3_wild", analysis="20200422", method="aim"):
        """Access species calls for one or more sample sets.

        Parameters
        ----------
        sample_sets : str or list of str
            Can be a sample set identifier (e.g., "AG1000G-AO") or a list of sample set
            identifiers (e.g., ["AG1000G-BF-A", "AG1000G-BF-B"] or a release identifier (e.g.,
            "v3") or a list of release identifiers.
        analysis : str
            Species calling analysis version.
        method : str
            Species calling method; "aim" is ancestry informative markers, "pca" is principal
            components analysis.

        Returns
        -------
        df : pandas.DataFrame

        """

        sample_sets = self._prep_sample_sets_arg(sample_sets=sample_sets)

        if isinstance(sample_sets, str):
            # assume single sample set
            df = self._read_species_calls(
                sample_set=sample_sets, analysis=analysis, method=method
            )

        else:
            # concatenate multiple sample sets
            dfs = [
                self.species_calls(sample_sets=c, analysis=analysis, method=method)
                for c in sample_sets
            ]
            df = pandas.concat(dfs, axis=0, sort=False).reset_index(drop=True)

        return df

    def sample_metadata(self, sample_sets="v3_wild", species_calls=("20200422", "aim")):
        """Access sample metadata for one or more sample sets.

        Parameters
        ----------
        sample_sets : str or list of str
            Can be a sample set identifier (e.g., "AG1000G-AO") or a list of sample set
            identifiers (e.g., ["AG1000G-BF-A", "AG1000G-BF-B"]) or a release identifier (e.g.,
            "v3") or a list of release identifiers.
        species_calls : (str, str), optional
            Include species calls in metadata.

        Returns
        -------
        df : pandas.DataFrame

        """

        sample_sets = self._prep_sample_sets_arg(sample_sets=sample_sets)

        if isinstance(sample_sets, str):
            # assume single sample set
            df = self._read_general_metadata(sample_set=sample_sets)
            if species_calls is not None:
                analysis, method = species_calls
                df_species = self._read_species_calls(
                    sample_set=sample_sets, analysis=analysis, method=method
                )
                df = df.merge(df_species, on="sample_id", sort=False)

        else:
            # concatenate multiple sample sets
            dfs = [
                self.sample_metadata(sample_sets=c, species_calls=species_calls)
                for c in sample_sets
            ]
            df = pandas.concat(dfs, axis=0, sort=False).reset_index(drop=True)

        return df

    def open_site_filters(self, mask, analysis="dt_20200416"):
        """Open site filters zarr.

        Parameters
        ----------
        mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Mask to use.
        analysis : str, optional
            Site filters analysis version.

        Returns
        -------
        genome : zarr.hierarchy.Group

        """
        key = mask, analysis
        try:
            return self._cache_site_filters[key]
        except KeyError:
            path = f"{self._path}/v3/site_filters/{analysis}/{mask}/"
            store = SafeStore(FSMap(root=path, fs=self._fs, check=False, create=False))
            root = zarr.open_consolidated(store=store)
            self._cache_site_filters[key] = root
            return root

    def site_filters(
        self,
        contig,
        mask,
        field="filter_pass",
        analysis="dt_20200416",
        inline_array=True,
        chunks="auto",
    ):
        """Access SNP site filters.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Mask to use.
        field : str, optional
            Array to access.
        analysis : str, optional
            Site filters analysis version.
        inline_array : bool, optional
            Passed through to dask.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.array.Array

        """

        root = self.open_site_filters(mask=mask, analysis=analysis)
        z = root[contig]["variants"][field]
        d = from_zarr(z, inline_array=inline_array, chunks=chunks)
        return d

    def open_snp_sites(self):
        """Open SNP sites zarr.

        Returns
        -------
        genome : zarr.hierarchy.Group

        """
        if self._cache_snp_sites is None:
            path = f"{self._path}/v3/snp_genotypes/all/sites/"
            store = SafeStore(FSMap(root=path, fs=self._fs, check=False, create=False))
            root = zarr.open_consolidated(store=store)
            self._cache_snp_sites = root
        return self._cache_snp_sites

    def snp_sites(
        self,
        contig,
        field=None,
        site_mask=None,
        site_filters="dt_20200416",
        inline_array=True,
        chunks="auto",
    ):
        """Access SNP site data (positions and alleles).

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        field : {"POS", "REF", "ALT"}, optional
            Array to access. If not provided, all three arrays POS, REF, ALT will be returned as a tuple.
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str
            Site filters analysis version.
        inline_array : bool, optional
            Passed through to dask.array.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.array.Array or tuple of dask.array.Array

        """

        if field is None:
            # return POS, REF, ALT
            ret = tuple(
                self.snp_sites(contig=contig, field=f, site_mask=None)
                for f in ("POS", "REF", "ALT")
            )

        else:
            root = self.open_snp_sites()
            z = root[contig]["variants"][field]
            ret = from_zarr(z, inline_array=inline_array, chunks=chunks)

        if site_mask is not None:
            loc_sites = self.site_filters(
                contig=contig, mask=site_mask, analysis=site_filters
            )
            if isinstance(ret, tuple):
                ret = tuple(dask_compress(loc_sites, d, axis=0) for d in ret)
            else:
                ret = dask_compress(loc_sites, ret, axis=0)

        return ret

    def open_snp_genotypes(self, sample_set):
        """Open SNP genotypes zarr.

        Parameters
        ----------
        sample_set : str

        Returns
        -------
        genome : zarr.hierarchy.Group

        """
        try:
            return self._cache_snp_genotypes[sample_set]
        except KeyError:
            release = self._lookup_release(sample_set=sample_set)
            path = f"{self._path}/{release}/snp_genotypes/all/{sample_set}/"
            store = SafeStore(FSMap(root=path, fs=self._fs, check=False, create=False))
            root = zarr.open_consolidated(store=store)
            self._cache_snp_genotypes[sample_set] = root
            return root

    def snp_genotypes(
        self,
        contig,
        sample_sets="v3_wild",
        field="GT",
        site_mask=None,
        site_filters="dt_20200416",
        inline_array=True,
        chunks="auto",
    ):
        """Access SNP genotypes and associated data.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        sample_sets : str or list of str
            Can be a sample set identifier (e.g., "AG1000G-AO") or a list of sample set
            identifiers (e.g., ["AG1000G-BF-A", "AG1000G-BF-B"]) or a release identifier (e.g.,
            "v3") or a list of release identifiers.
        field : {"GT", "GQ", "AD", "MQ"}
            Array to access.
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str, optional
            Site filters analysis version.
        inline_array : bool, optional
            Passed through to dask.array.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.array.Array

        """

        sample_sets = self._prep_sample_sets_arg(sample_sets=sample_sets)

        if isinstance(sample_sets, str):
            # single sample set
            root = self.open_snp_genotypes(sample_set=sample_sets)
            z = root[contig]["calldata"][field]
            d = from_zarr(z, inline_array=inline_array, chunks=chunks)

        else:
            # concatenate multiple sample sets
            ds = [
                self.snp_genotypes(contig=contig, sample_sets=c, field=field)
                for c in sample_sets
            ]
            d = da.concatenate(ds, axis=1)

        if site_mask is not None:
            loc_sites = self.site_filters(
                contig=contig, mask=site_mask, analysis=site_filters
            )
            d = dask_compress(loc_sites, d, axis=0)

        return d

    def open_genome(self):
        """Open the reference genome zarr.

        Returns
        -------
        genome : zarr.hierarchy.Group

        """
        if self._cache_genome is None:
            path = f"{self._path}/reference/genome/agamp4/Anopheles-gambiae-PEST_CHROMOSOMES_AgamP4.zarr"
            store = SafeStore(FSMap(root=path, fs=self._fs, check=False, create=False))
            self._cache_genome = zarr.open_consolidated(store=store)
        return self._cache_genome

    def genome_sequence(self, contig, inline_array=True, chunks="auto"):
        """Access the reference genome sequence.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        inline_array : bool, optional
            Passed through to dask.array.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.array.Array

        """
        genome = self.open_genome()
        z = genome[contig]
        d = from_zarr(z, inline_array=inline_array, chunks=chunks)
        return d

    def geneset(self, attributes=("ID", "Parent", "Name")):
        """Access genome feature annotations (AgamP4.12).

        Parameters
        ----------
        attributes : list of str, optional
            Attribute keys to unpack into columns. Provide "*" to unpack all attributes.

        Returns
        -------
        df : pandas.DataFrame

        """

        if attributes is not None:
            attributes = tuple(attributes)

        try:
            df = self._cache_geneset[attributes]

        except KeyError:
            path = f"{self._path}/{gff3_path}"
            with self._fs.open(path, mode="rb") as f:
                df = read_gff3(f, compression="gzip")
            if attributes is not None:
                df = unpack_gff3_attributes(df, attributes=attributes)
            self._cache_geneset[attributes] = df

        return df

    def is_accessible(self, contig, site_mask, site_filters="dt_20200416"):
        """Compute genome accessibility array.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str, optional
            Site filters analysis version.

        Returns
        -------
        a : numpy.ndarray

        """

        # determine contig sequence length
        seq_length = self.genome_sequence(contig).shape[0]

        # setup output
        is_accessible = np.zeros(seq_length, dtype=bool)

        # access positions
        pos = self.snp_sites(contig, field="POS").compute()

        # access site filters
        filter_pass = self.site_filters(
            contig, mask=site_mask, analysis=site_filters
        ).compute()

        # assign values from site filters
        is_accessible[pos - 1] = filter_pass

        return is_accessible

    def snp_effects(self, transcript, site_mask):
        """Compute variant effects for a gene transcript.

        Parameters
        ----------
        transcript : str
            Gene transcript ID (AgamP4.12), e.g., "AGAP004707-RA".
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.

        Returns
        -------
        df : pandas.DataFrame

        """
        # take an AGAP transcript ID and get meta data from the gff using veff
        # first time sets up and caches ann object
        if self._cache_annotator is None:
            self._cache_annotator = veff.Annotator(
                genome=self.open_genome(), geneset=self.geneset()
            )

        ann = self._cache_annotator
        feature = ann.get_feature(transcript)
        contig = feature[0]
        start = feature[3]
        stop = feature[4]

        # grab pos, ref and alt for chrom arm from snp_sites
        sites = self.snp_sites(contig=contig, site_mask=site_mask)

        # sites are dask arrays, turn pos into sorted index
        pos = allel.SortedIndex(sites[0].compute())
        # locate transcript range
        loc = pos.locate_range(start, stop)
        # dask compute on the sliced arrays to speed things up
        ref = sites[1][loc].compute()
        alt = sites[2][loc].compute()

        # build an initial dataframe with contig, pos, ref, alt columns
        df_in = pandas.DataFrame()
        df_in["position"] = np.asarray(pos[loc])
        df_in["ref_allele"] = [q.tobytes().decode() for q in np.asarray(ref)]
        # bytes within lists within lists...
        df_in["alt_allele"] = [list(q.tobytes().decode()) for q in list(alt)]
        # explode the alt alleles into their own rows
        df_effects = df_in.explode("alt_allele").reset_index(drop=True)

        # then, iterate over rows of the dataframe, calling get_effects()
        # for each row, and using that to build additional columns effect,
        # impact, etc.

        leffect = []
        limpact = []
        lref_codon = []
        lalt_codon = []
        laa_pos = []
        lref_aa = []
        lalt_aa = []
        laa_change = []

        for row in df_effects.itertuples(index=True):
            for effect in ann.get_effects(
                chrom=contig,
                pos=row.position,
                ref=row.ref_allele,
                alt=row.alt_allele,
                transcript_ids=[transcript],
            ):
                leffect.append(effect.effect)
                limpact.append(effect.impact)
                lref_codon.append(effect.ref_codon)
                lalt_codon.append(effect.alt_codon)
                laa_pos.append(effect.aa_pos)
                lref_aa.append(effect.ref_aa)
                lalt_aa.append(effect.alt_aa)
                laa_change.append(effect.aa_change)

        df_effects["effect"] = leffect
        df_effects["impact"] = limpact
        df_effects["ref_codon"] = lref_codon
        df_effects["alt_codon"] = lalt_codon
        df_effects["aa_pos"] = laa_pos
        df_effects["ref_aa"] = lref_aa
        df_effects["alt_aa"] = lalt_aa
        df_effects["aa_change"] = laa_change

        return df_effects

    def snp_allele_frequencies(
        self,
        transcript,
        populations,
        site_mask=None,
        site_filters="dt_20200416",
        species_calls=("20200422", "aim"),
        sample_sets="v3_wild",
        drop_invariant=True,
    ):
        """Compute per variant population allele frequencies for a gene transcript.

        Parameters
        ----------
        transcript : str
            Gene transcript ID (AgamP4.12), e.g., "AGAP004707-RA".
        populations : dict
            Dictionary to map population IDs to sample queries, e.g.,
            "bf_2012_col": "country == 'Burkina Faso' and year == 2012 and species == 'coluzzii'"
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str, optional
            Site filters analysis version.
        species_calls : (str, str), optional
            Include species calls in metadata.
        sample_sets : str or list of str, optional
            Can be a sample set identifier (e.g., "AG1000G-AO") or a list of sample set
            identifiers (e.g., ["AG1000G-BF-A", "AG1000G-BF-B"]) or a release identifier (e.g.,
            "v3") or a list of release identifiers.
        drop_invariant : bool, optional
            If True, variants with no alternate allele calls in any populations are dropped from the result.

        Returns
        -------
        df : pandas.DataFrame

        """
        # get feature details (don't need to set up an annotator here)
        geneset = self.geneset()
        geneset = geneset.set_index("ID")
        feature = geneset.loc[transcript]
        contig = feature.seqid
        start = feature.start
        stop = feature.end

        # grab pos, ref and alt for chrom arm from snp_sites
        pos, ref, alt = self.snp_sites(contig=contig, site_mask=site_mask)

        # sites are dask arrays, turn pos into sorted index
        pos = allel.SortedIndex(pos.compute())
        # locate transcript range
        loc = pos.locate_range(start, stop)

        # we want to grab all metadata then get idx for samples we want
        df_meta = self.sample_metadata(
            sample_sets=sample_sets, species_calls=species_calls
        )

        # get genotypes
        gt = self.snp_genotypes(
            contig=contig,
            sample_sets=sample_sets,
            field="GT",
            site_mask=site_mask,
            site_filters=site_filters,
        )

        # chop everything to loc
        gt = gt[loc].compute()
        pos = pos[loc]
        ref = ref[loc].compute()
        alt = alt[loc].compute()

        # count alleles
        afs = dict()
        for pop, query in populations.items():
            loc_pop = df_meta.eval(query).values
            gt_pop = da.compress(loc_pop, gt, axis=1)
            ac_pop = (
                allel.GenotypeDaskArray(gt_pop).count_alleles(max_allele=3).compute()
            )
            afs[pop] = ac_pop.to_frequencies()

        # set up columns
        cols = {
            "position": np.repeat(pos, 3),
            "ref_allele": np.repeat(ref.astype("U1"), 3),
            "alt_allele": alt.astype("U1").flatten(),
        }

        for pop in populations:
            cols[pop] = afs[pop][:, 1:].flatten()

        # build df
        df = pandas.DataFrame(cols)

        # drop invariants
        if drop_invariant:
            loc_variant = df[populations].sum(axis=1) > 0
            df = df[loc_variant]

        # add max freq column
        df["maximum"] = df[populations].max(axis=1)

        return df

    def cross_metadata(self):
        """Load a dataframe containing metadata about samples in colony crosses, including
        which samples are parents or progeny in which crosses.

        Returns
        -------
        df : pandas.DataFrame

        """

        if self._cache_cross_metadata is None:

            path = f"{self._path}/v3/metadata/crosses/crosses.fam"
            fam_names = [
                "cross",
                "sample_id",
                "father_id",
                "mother_id",
                "sex",
                "phenotype",
            ]
            with self._fs.open(path) as f:
                df = pandas.read_csv(
                    f,
                    sep="\t",
                    na_values=["", "0"],
                    names=fam_names,
                    dtype={"sex": str},
                )

            # convert "sex" column for consistency with sample metadata
            df.loc[df["sex"] == "1", "sex"] = "M"
            df.loc[df["sex"] == "2", "sex"] = "F"

            # add a "role" column for convenience
            df["role"] = "progeny"
            df.loc[df["mother_id"].isna(), "role"] = "parent"

            # drop "phenotype" column, not used
            df.drop("phenotype", axis="columns", inplace=True)

            self._cache_cross_metadata = df

        return self._cache_cross_metadata

    def open_site_annotations(self):
        if self._cache_site_annotations is None:
            path = f"{self._path}/reference/genome/agamp4/Anopheles-gambiae-PEST_SEQANNOTATION_AgamP4.12.zarr"
            store = SafeStore(FSMap(root=path, fs=self._fs, check=False, create=False))
            self._cache_site_annotations = zarr.open_consolidated(store=store)
        return self._cache_site_annotations

    def site_annotations(
        self,
        contig,
        field,
        site_mask=None,
        site_filters="dt_20200416",
        inline_array=True,
        chunks="auto",
    ):
        """Load site annotations.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        field : str
            One of "codon_degeneracy", "codon_nonsyn", "codon_position", "seq_cls",
            "seq_flen", "seq_relpos_start", "seq_relpos_stop".
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str
            Site filters analysis version.
        inline_array : bool, optional
            Passed through to dask.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        d : dask.Array

        """

        # access the array of values for all genome positions
        root = self.open_site_annotations()
        d = from_zarr(root[field][contig], inline_array=inline_array, chunks=chunks)

        # access and subset to SNP positions
        pos = self.snp_sites(
            contig=contig, field="POS", site_mask=site_mask, site_filters=site_filters
        )
        d = da.take(d, pos - 1)

        return d

    def _snp_calls_dataset(
        self, contig, sample_set, site_filters, inline_array, chunks
    ):

        coords = dict()
        data_vars = dict()

        # variant arrays
        sites_root = self.open_snp_sites()

        # variant_position
        pos_z = sites_root[contig]["variants"]["POS"]
        variant_position = from_zarr(pos_z, inline_array=inline_array, chunks=chunks)
        coords["variant_position"] = [DIM_VARIANT], variant_position

        # variant_allele
        ref_z = sites_root[contig]["variants"]["REF"]
        ref = from_zarr(ref_z, inline_array=inline_array, chunks=chunks)
        alt_z = sites_root[contig]["variants"]["ALT"]
        alt = from_zarr(alt_z, inline_array=inline_array, chunks=chunks)
        variant_allele = da.concatenate([ref[:, None], alt], axis=1)
        data_vars["variant_allele"] = [DIM_VARIANT, DIM_ALLELE], variant_allele

        # variant_contig
        contig_index = self.contigs.index(contig)
        variant_contig = da.full_like(
            variant_position, fill_value=contig_index, dtype="u1"
        )
        coords["variant_contig"] = [DIM_VARIANT], variant_contig

        # site filters arrays
        for mask in "gamb_colu_arab", "gamb_colu", "arab":
            filters_root = self.open_site_filters(mask=mask, analysis=site_filters)
            z = filters_root[contig]["variants"]["filter_pass"]
            d = from_zarr(z, inline_array=inline_array, chunks=chunks)
            data_vars[f"variant_filter_pass_{mask}"] = [DIM_VARIANT], d

        # call arrays
        calls_root = self.open_snp_genotypes(sample_set=sample_set)
        gt_z = calls_root[contig]["calldata"]["GT"]
        call_genotype = from_zarr(gt_z, inline_array=inline_array, chunks=chunks)
        gq_z = calls_root[contig]["calldata"]["GQ"]
        call_gq = from_zarr(gq_z, inline_array=inline_array, chunks=chunks)
        ad_z = calls_root[contig]["calldata"]["AD"]
        call_ad = from_zarr(ad_z, inline_array=inline_array, chunks=chunks)
        mq_z = calls_root[contig]["calldata"]["MQ"]
        call_mq = from_zarr(mq_z, inline_array=inline_array, chunks=chunks)
        data_vars["call_genotype"] = (
            [DIM_VARIANT, DIM_SAMPLE, DIM_PLOIDY],
            call_genotype,
        )
        data_vars["call_GQ"] = ([DIM_VARIANT, DIM_SAMPLE], call_gq)
        data_vars["call_MQ"] = ([DIM_VARIANT, DIM_SAMPLE], call_mq)
        data_vars["call_AD"] = ([DIM_VARIANT, DIM_SAMPLE, DIM_ALLELE], call_ad)

        # sample arrays
        z = calls_root["samples"]
        sample_id = from_zarr(z, inline_array=inline_array, chunks=chunks)
        coords["sample_id"] = [DIM_SAMPLE], sample_id

        # setup attributes
        attrs = {"contigs": self.contigs}

        # create a dataset
        ds = xarray.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)

        return ds

    def snp_calls(
        self,
        contig,
        sample_sets="v3_wild",
        site_mask=None,
        site_filters="dt_20200416",
        inline_array=True,
        chunks="auto",
    ):
        """Access SNP sites, site filters and genotype calls.

        Parameters
        ----------
        contig : str
            Chromosome arm, e.g., "3R".
        sample_sets : str or list of str
            Can be a sample set identifier (e.g., "AG1000G-AO") or a list of sample set
            identifiers (e.g., ["AG1000G-BF-A", "AG1000G-BF-B"]) or a release identifier (e.g.,
            "v3") or a list of release identifiers.
        site_mask : {"gamb_colu_arab", "gamb_colu", "arab"}
            Site filters mask to apply.
        site_filters : str
            Site filters analysis version.
        inline_array : bool, optional
            Passed through to dask.array.from_array().
        chunks : str, optional
            If 'auto' let dask decide chunk size. If 'native' use native zarr chunks.
            Also can be a target size, e.g., '200 MiB'.

        Returns
        -------
        ds : xarray.Dataset

        """

        sample_sets = self._prep_sample_sets_arg(sample_sets=sample_sets)

        if isinstance(sample_sets, str):

            # single sample set requested
            ds = self._snp_calls_dataset(
                contig=contig,
                sample_set=sample_sets,
                site_filters=site_filters,
                inline_array=inline_array,
                chunks=chunks,
            )

        else:

            # multiple sample sets requested, need to concatenate along samples dimension
            datasets = [
                self._snp_calls_dataset(
                    contig=contig,
                    sample_set=sample_set,
                    site_filters=site_filters,
                    inline_array=inline_array,
                    chunks=chunks,
                )
                for sample_set in sample_sets
            ]
            ds = xarray.concat(
                datasets,
                dim=DIM_SAMPLE,
                data_vars="minimal",
                coords="minimal",
                compat="override",
                join="override",
            )

        # apply site filters
        if site_mask is not None:
            ds = dask_compress_dataset(
                ds, indexer=f"variant_filter_pass_{site_mask}", dim=DIM_VARIANT
            )

        # add call_genotype_mask
        ds["call_genotype_mask"] = ds["call_genotype"] < 0

        return ds

    def snp_dataset(self, *args, **kwargs):
        # backwards compatibility, this method has been renamed to snp_calls()
        return self.snp_calls(*args, **kwargs)
